# -*- coding: utf8 -*-
import datetime

from unittest import TestCase

import pytz

from freezegun import freeze_time
from mock import Mock, patch

from helga.plugins import reminders


class DoReminderTestCase(TestCase):

    def setUp(self):
        reminders._scheduled.add(1)
        self.rec = {'channel': '#bots', 'message': 'some message'}
        self.now = datetime.datetime(day=11, month=12, year=2013)  # A wednesday
        self.client = Mock()

    @patch('helga.plugins.reminders.db')
    def test_do_reminder_simple(self, db):
        db.reminders.find_one.return_value = self.rec
        reminders._do_reminder(1, self.client)

        assert 1 not in reminders._scheduled
        db.reminders.remove.assert_called_with(1)
        self.client.msg.assert_called_with('#bots', 'some message')

    @patch('helga.plugins.reminders.db')
    @patch('helga.plugins.reminders.reactor')
    def test_do_reminder_reschedules_same_week(self, reactor, db):
        self.rec['when'] = self.now
        self.rec['repeat'] = [0, 2, 4]  # M, W, F
        db.reminders.find_one.return_value = self.rec

        with freeze_time(self.now):
            reminders._do_reminder(1, self.client)

        rec_upd = self.rec.copy()
        rec_upd['when'] = datetime.datetime(day=13, month=12, year=2013)

        db.reminders.save.assert_called_with(rec_upd)
        reactor.callLater.assert_called_with(48 * 3600, reminders._do_reminder, 1, self.client)

    @patch('helga.plugins.reminders.db')
    @patch('helga.plugins.reminders.reactor')
    def test_do_reminder_reschedules_next_week(self, reactor, db):
        self.rec['when'] = datetime.datetime(day=13, month=12, year=2013)
        self.rec['repeat'] = [0, 2, 4]  # M, W, F
        db.reminders.find_one.return_value = self.rec

        with freeze_time(self.rec['when']):
            reminders._do_reminder(1, self.client)

        rec_upd = self.rec.copy()
        rec_upd['when'] = datetime.datetime(day=16, month=12, year=2013)

        db.reminders.save.assert_called_with(rec_upd)
        reactor.callLater.assert_called_with(72 * 3600, reminders._do_reminder, 1, self.client)

    @patch('helga.plugins.reminders._scheduled')
    @patch('helga.plugins.reminders.db')
    def test_scheduled_discarded_with_no_record(self, db, scheduled):
        db.reminders.find_one.return_value = None
        reminders._do_reminder(1, Mock())
        scheduled.discard.assert_called_with(1)

    @patch('helga.plugins.reminders.db')
    def test_handles_unicode(self, db):
        client = Mock()
        snowman = u'☃'
        reminder = {
            'channel': snowman,
            'message': snowman,
        }
        db.reminders.find_one.return_value = reminder
        reminders._do_reminder(1, client)
        client.msg.assert_called_with(snowman, snowman)


class InReminderTestCase(TestCase):

    def setUp(self):
        reminders._scheduled.clear()
        self.client = Mock()
        self.now = datetime.datetime(day=13, month=12, year=2013)

    @patch('helga.plugins.reminders.db')
    @patch('helga.plugins.reminders.reactor')
    def test_in_reminder_for_different_channel(self, reactor, db):
        db.reminders.insert.return_value = 1

        with freeze_time(self.now):
            reminders.in_reminder(self.client, '#bots', 'me',
                                  ['12m', 'on', '#foo', 'this', 'is', 'the', 'message'])

        inserted = db.reminders.insert.call_args[0][0]
        assert inserted['channel'] == '#foo'
        assert inserted['message'] == 'this is the message'

    @patch('helga.plugins.reminders.db')
    @patch('helga.plugins.reminders.reactor')
    def test_in_reminder_for_different_channel_adds_chan_hash(self, reactor, db):
        db.reminders.insert.return_value = 1

        with freeze_time(self.now):
            reminders.in_reminder(self.client, '#bots', 'me',
                                  ['12m', 'on', 'foo', 'this', 'is', 'the', 'message'])

        inserted = db.reminders.insert.call_args[0][0]
        assert inserted['channel'] == '#foo'
        assert inserted['message'] == 'this is the message'

    @patch('helga.plugins.reminders.db')
    @patch('helga.plugins.reminders.reactor')
    def test_in_reminder_for_minutes(self, reactor, db):
        db.reminders.insert.return_value = 1

        with freeze_time(self.now):
            reminders.in_reminder(self.client, '#bots', 'me', ['12m', 'this', 'is', 'the', 'message'])

        inserted = db.reminders.insert.call_args[0][0]

        assert inserted['message'] == 'this is the message'
        assert inserted['channel'] == '#bots'
        assert reactor.callLater.call_args[0] == (12 * 60, reminders._do_reminder, 1, self.client)

    @patch('helga.plugins.reminders.db')
    @patch('helga.plugins.reminders.reactor')
    def test_in_reminder_for_hours(self, reactor, db):
        db.reminders.insert.return_value = 1

        with freeze_time(self.now):
            reminders.in_reminder(self.client, '#bots', 'me', ['12h', 'this', 'is', 'the', 'message'])

        inserted = db.reminders.insert.call_args[0][0]

        assert inserted['message'] == 'this is the message'
        assert inserted['channel'] == '#bots'
        assert reactor.callLater.call_args[0] == (12 * 3600, reminders._do_reminder, 1, self.client)

    @patch('helga.plugins.reminders.db')
    @patch('helga.plugins.reminders.reactor')
    def test_in_reminder_for_days(self, reactor, db):
        db.reminders.insert.return_value = 1

        with freeze_time(self.now):
            reminders.in_reminder(self.client, '#bots', 'me', ['12d', 'this', 'is', 'the', 'message'])

        inserted = db.reminders.insert.call_args[0][0]

        assert inserted['message'] == 'this is the message'
        assert inserted['channel'] == '#bots'
        assert reactor.callLater.call_args[0] == (12 * 24 * 3600, reminders._do_reminder, 1, self.client)

    def test_in_reminder_for_unknown(self):
        resp = reminders.in_reminder(self.client, '#bots', 'me', ['12x', 'this', 'is', 'the', 'message'])
        assert resp.startswith("Sorry I didn't understand '12x'")


class AtReminderTestCase(TestCase):

    def setUp(self):
        self.client = Mock()
        self.now = datetime.datetime(day=11, month=12, year=2013, hour=12)
        self.past = datetime.datetime(day=9, month=12, year=2013, hour=12)
        self.future = datetime.datetime(day=13, month=12, year=2013, hour=12)

        self.tz = pytz.timezone('US/Eastern')

        reminders._scheduled.clear()

    @patch('helga.plugins.reminders.db')
    @patch('helga.plugins.reminders.reactor')
    def test_using_different_channel_and_with_repeat(self, reactor, db):
        args = ['13:00', 'on', '#foo', 'test', 'message', 'repeat', 'MWF']
        db.reminders.insert.return_value = 1

        # Account for UTC difference
        with freeze_time(self.now + datetime.timedelta(hours=5)):
            reminders.at_reminder(self.client, '#bots', 'me', args)

        rec = db.reminders.insert.call_args[0][0]
        assert rec['channel'] == '#foo'
        assert rec['message'] == 'test message'

    @patch('helga.plugins.reminders.db')
    @patch('helga.plugins.reminders.reactor')
    def test_using_different_channel(self, reactor, db):
        args = ['13:00', 'on', '#foo', 'this is a message']
        db.reminders.insert.return_value = 1

        # Account for UTC difference
        with freeze_time(self.now + datetime.timedelta(hours=5)):
            reminders.at_reminder(self.client, '#bots', 'me', args)

        rec = db.reminders.insert.call_args[0][0]
        assert rec['channel'] == '#foo'
        assert rec['message'] == 'this is a message'

    @patch('helga.plugins.reminders.db')
    @patch('helga.plugins.reminders.reactor')
    def test_using_different_channel_when_timezone_present(self, reactor, db):
        args = ['13:00', 'EST', 'on', '#foo', 'this is a message']
        db.reminders.insert.return_value = 1

        # Account for UTC difference
        with freeze_time(self.now + datetime.timedelta(hours=5)):
            reminders.at_reminder(self.client, '#bots', 'me', args)

        rec = db.reminders.insert.call_args[0][0]
        assert rec['channel'] == '#foo'
        assert rec['message'] == 'this is a message'

    @patch('helga.plugins.reminders.db')
    @patch('helga.plugins.reminders.reactor')
    def test_using_different_channel_adds_chan_hash(self, reactor, db):
        args = ['13:00', 'on', 'foo', 'this is a message']
        db.reminders.insert.return_value = 1

        # Account for UTC difference
        with freeze_time(self.now + datetime.timedelta(hours=5)):
            reminders.at_reminder(self.client, '#bots', 'me', args)

        rec = db.reminders.insert.call_args[0][0]
        assert rec['channel'] == '#foo'
        assert rec['message'] == 'this is a message'

    @patch('helga.plugins.reminders.db')
    @patch('helga.plugins.reminders.reactor')
    def test_no_tz_no_repeat_in_future(self, reactor, db):
        args = ['13:00', 'this is a message']
        db.reminders.insert.return_value = 1

        # Account for UTC difference
        with freeze_time(self.now + datetime.timedelta(hours=5)):
            reminders.at_reminder(self.client, '#bots', 'me', args)

        rec = db.reminders.insert.call_args[0][0]
        expect = self.now + datetime.timedelta(hours=1)

        # 'when' will be stored as UTC timestamp
        when = rec['when'].astimezone(self.tz).replace(tzinfo=None)

        assert 'repeat' not in rec
        assert when == expect
        assert rec['channel'] == '#bots'
        assert rec['message'] == 'this is a message'
        reactor.callLater.assert_called_with(1*3600, reminders._do_reminder, 1, self.client)

    @patch('helga.plugins.reminders.db')
    @patch('helga.plugins.reminders.reactor')
    def test_no_tz_no_repeat_in_past(self, reactor, db):
        args = ['6:00', 'this is a message']
        db.reminders.insert.return_value = 1

        # Account for UTC difference
        with freeze_time(self.now + datetime.timedelta(hours=5)):
            reminders.at_reminder(self.client, '#bots', 'me', args)

        rec = db.reminders.insert.call_args[0][0]
        expect = self.now + datetime.timedelta(hours=18)

        # 'when' will be stored as UTC timestamp
        when = rec['when'].astimezone(self.tz).replace(tzinfo=None)

        assert 'repeat' not in rec
        assert when == expect
        assert rec['channel'] == '#bots'
        assert rec['message'] == 'this is a message'
        reactor.callLater.assert_called_with(18*3600, reminders._do_reminder, 1, self.client)

    @patch('helga.plugins.reminders.db')
    @patch('helga.plugins.reminders.reactor')
    def test_tz_no_repeat_in_future(self, reactor, db):
        args = ['13:00', 'US/Central', 'this is a message']
        db.reminders.insert.return_value = 1

        # Account for UTC difference
        with freeze_time(self.now + datetime.timedelta(hours=6)):
            reminders.at_reminder(self.client, '#bots', 'me', args)

        rec = db.reminders.insert.call_args[0][0]
        expect = self.now + datetime.timedelta(hours=1)

        # 'when' will be stored as UTC timestamp
        when = rec['when'].astimezone(pytz.timezone('US/Central')).replace(tzinfo=None)

        assert 'repeat' not in rec
        assert when == expect
        assert rec['channel'] == '#bots'
        assert rec['message'] == 'this is a message'
        reactor.callLater.assert_called_with(1*3600, reminders._do_reminder, 1, self.client)

    @patch('helga.plugins.reminders.db')
    @patch('helga.plugins.reminders.reactor')
    def test_tz_no_repeat_in_past(self, reactor, db):
        args = ['6:00', 'US/Central', 'this is a message']
        db.reminders.insert.return_value = 1

        # Account for UTC difference
        with freeze_time(self.now + datetime.timedelta(hours=6)):
            reminders.at_reminder(self.client, '#bots', 'me', args)

        rec = db.reminders.insert.call_args[0][0]
        expect = self.now + datetime.timedelta(hours=18)

        # 'when' will be stored as UTC timestamp
        when = rec['when'].astimezone(pytz.timezone('US/Central')).replace(tzinfo=None)

        assert 'repeat' not in rec
        assert when == expect
        assert rec['channel'] == '#bots'
        assert rec['message'] == 'this is a message'
        reactor.callLater.assert_called_with(18*3600, reminders._do_reminder, 1, self.client)

    @patch('helga.plugins.reminders.db')
    @patch('helga.plugins.reminders.reactor')
    def test_tz_repeat_in_future(self, reactor, db):
        args = ['13:00', 'US/Central', 'this is a message', 'repeat', 'MWF']
        db.reminders.insert.return_value = 1

        # Account for UTC difference
        with freeze_time(self.now + datetime.timedelta(hours=6)):
            reminders.at_reminder(self.client, '#bots', 'me', args)

        rec = db.reminders.insert.call_args[0][0]
        expect = self.now + datetime.timedelta(hours=1)

        # 'when' will be stored as UTC timestamp
        when = rec['when'].astimezone(pytz.timezone('US/Central')).replace(tzinfo=None)

        assert when == expect
        assert rec['channel'] == '#bots'
        assert rec['message'] == 'this is a message'
        assert rec['repeat'] == [0, 2, 4]
        reactor.callLater.assert_called_with(1*3600, reminders._do_reminder, 1, self.client)

    @patch('helga.plugins.reminders.db')
    @patch('helga.plugins.reminders.reactor')
    def test_tz_repeat_in_past(self, reactor, db):
        args = ['6:00', 'US/Central', 'this is a message', 'repeat', 'MWF']
        db.reminders.insert.return_value = 1

        # Account for UTC difference
        with freeze_time(self.now + datetime.timedelta(hours=6)):
            reminders.at_reminder(self.client, '#bots', 'me', args)

        rec = db.reminders.insert.call_args[0][0]
        expect = self.now + datetime.timedelta(hours=42)  # We expect it to be on Friday now

        # 'when' will be stored as UTC timestamp
        when = rec['when'].astimezone(pytz.timezone('US/Central')).replace(tzinfo=None)

        assert when == expect
        assert rec['channel'] == '#bots'
        assert rec['message'] == 'this is a message'
        assert rec['repeat'] == [0, 2, 4]
        reactor.callLater.assert_called_with(42*3600, reminders._do_reminder, 1, self.client)

    @patch('helga.plugins.reminders.db')
    @patch('helga.plugins.reminders.reactor')
    def test_invalid_days_returns_warning(self, reactor, db):
        # Invalid chars
        args = ['6:00', 'US/Central', 'this is a message', 'repeat', 'XYZ']
        assert "I didn't understand" in reminders.at_reminder(self.client, '#bots', 'me', args)
        assert not db.reminders.insert.called
        assert not reactor.callLater.called

        # No chars
        args = ['6:00', 'US/Central', 'this is a message', 'repeat', '']
        assert "I didn't understand" in reminders.at_reminder(self.client, '#bots', 'me', args)
        assert not db.reminders.insert.called
        assert not reactor.callLater.called


class ReadableTimeTestCase(TestCase):

    def test_readable_time_delta_minutes_only(self):
        ret = reminders.readable_time_delta(610)
        assert ret == '10 minutes'

    def test_readable_time_delta_hours_and_minutes(self):
        ret = reminders.readable_time_delta((3 * 3600) + 610)
        assert ret == '3 hours and 10 minutes'

    def test_readable_time_delta_days_hours_and_minutes(self):
        ret = reminders.readable_time_delta((8 * 86400) + (3 * 3600) + 610)
        assert ret == '8 days, 3 hours and 10 minutes'

    def test_readable_time_delta_singular_minutes(self):
        ret = reminders.readable_time_delta(65)
        assert ret == '1 minute'

    def test_readable_time_delta_singular_hours_and_minutes(self):
        ret = reminders.readable_time_delta((1 * 3600) + 65)
        assert ret == '1 hour and 1 minute'

    def test_readable_time_delta_singular_days_hours_and_minutes(self):
        ret = reminders.readable_time_delta((1 * 86400) + (1 * 3600) + 65)
        assert ret == '1 day, 1 hour and 1 minute'


class ListRemindersTestCase(TestCase):

    def setUp(self):
        self.rec = {
            '_id': '1234567890abcdefg',
            'when': datetime.datetime(year=2013, month=12, day=11, hour=13, minute=15, tzinfo=pytz.UTC),
            'message': 'Standup Time!',
        }

    @patch('helga.plugins.reminders.list_reminders')
    def test_list_reponds_via_privmsg(self, list_reminders):
        client = Mock()

        assert reminders.reminders(client, '#all', 'sduncan', 'reminders list', 'reminders', ['list']) is None
        client.me.assert_called_with('#all', 'whispers to sduncan')
        list_reminders.assert_called_with(client, 'sduncan', '#all')

    @patch('helga.plugins.reminders.list_reminders')
    def test_list_reponds_via_privmsg_for_specific_chan(self, list_reminders):
        client = Mock()

        assert reminders.reminders(client, '#all', 'sduncan', 'reminders list #bots',
                                   'reminders', ['list', '#bots']) is None
        client.me.assert_called_with('#all', 'whispers to sduncan')
        list_reminders.assert_called_with(client, 'sduncan', '#bots')

    @patch('helga.plugins.reminders.db')
    def test_list_no_results(self, db):
        client = Mock()
        db.reminders.find.return_value = []
        reminders.list_reminders(client, 'sduncan', '#bots')

        client.msg.assert_called_with('sduncan', "There are no reminders for channel: #bots")

    @patch('helga.plugins.reminders.db')
    def test_simple(self, db):
        client = Mock()
        client.msg = client

        db.reminders.find.return_value = [self.rec]
        reminders.list_reminders(client, 'sduncan', '#bots')

        client.msg.assert_called_with('sduncan',
                                      "sduncan, here are the reminders for channel: #bots\n"
                                      "[123456] At 12/11/13 13:15 UTC: 'Standup Time!'")

    @patch('helga.plugins.reminders.db')
    def test_with_repeats(self, db):
        client = Mock()
        client.msg = client

        self.rec['repeat'] = [0, 2, 4]
        db.reminders.find.return_value = [self.rec]
        reminders.list_reminders(client, 'sduncan', '#bots')

        client.msg.assert_called_with('sduncan',
                                      "sduncan, here are the reminders for channel: #bots\n"
                                      "[123456] At 12/11/13 13:15 UTC: 'Standup Time!' (Repeat every M,W,F)")


class InitRemindersTestCase(TestCase):

    @patch('helga.plugins.reminders.db')
    def test_updates_hash(self, db):
        records = [
            {
                'hash': 'foo',
                '_id': 1234567890,
                'when': datetime.datetime(day=13, month=12, year=2013)
            }
        ]
        db.reminders.find.return_value = records

        reminders.init_reminders(Mock())
        db.reminders.update.assert_called_with({'_id': 1234567890},
                                               {'$set': {'hash': '123456'}})

    @patch('helga.plugins.reminders.reactor')
    @patch('helga.plugins.reminders.db')
    def test_ignores_scheduled_reminder(self, db, reactor):
        records = [
            {
                'hash': 'foo',
                '_id': 1234567890,
                'when': datetime.datetime(day=13, month=12, year=2013)
            }
        ]
        db.reminders.find.return_value = records

        reminders._scheduled = set([123456789])
        reminders.init_reminders(Mock())
        assert not reactor.callLater.called

    @patch('helga.plugins.reminders.reactor')
    @patch('helga.plugins.reminders.db')
    def test_schedules_reminder(self, db, reactor):
        records = [
            {
                'hash': 'foo',
                '_id': 1234567890,
                'when': datetime.datetime(day=13, month=12, year=2013, tzinfo=pytz.UTC)
            }
        ]
        db.reminders.find.return_value = records

        with freeze_time(records[0]['when']):
            client = Mock()
            reminders._scheduled = set()
            reminders.init_reminders(client)
            assert 1234567890 in reminders._scheduled
            reactor.callLater.assert_called_with(0, reminders._do_reminder, 1234567890, client)

    @patch('helga.plugins.reminders.db')
    def test_with_stale_reminder(self, db):
        records = [
            {
                'hash': 'foo',
                '_id': 1234567890,
                'when': datetime.datetime(day=13, month=12, year=2013)
            }
        ]
        db.reminders.find.return_value = records

        with freeze_time(records[0]['when'] + datetime.timedelta(days=1)):
            client = Mock()
            reminders._scheduled = set()
            reminders.init_reminders(client)
            assert 1234567890 not in reminders._scheduled
            db.reminders.remove.assert_called_with(1234567890)

    @patch('helga.plugins.reminders.reactor')
    @patch('helga.plugins.reminders.db')
    def test_with_late_reminder(self, db, reactor):
        records = [
            {
                'hash': 'foo',
                '_id': 1234567890,
                'when': datetime.datetime(day=13, month=12, year=2013)
            }
        ]
        db.reminders.find.return_value = records

        with freeze_time(records[0]['when'] + datetime.timedelta(seconds=60)):
            client = Mock()
            reminders._scheduled = set()
            reminders.init_reminders(client)
            assert 1234567890 in reminders._scheduled
            reactor.callLater.assert_called_with(0, reminders._do_reminder, 1234567890, client)

    @patch('helga.plugins.reminders.reactor')
    @patch('helga.plugins.reminders.db')
    def test_with_repeated_reminder(self, db, reactor):
        records = [
            {
                'hash': 'foo',
                '_id': 1234567890,
                'when': datetime.datetime(day=13, month=12, year=2013),
                'repeat': range(7),
            }
        ]
        db.reminders.find.return_value = records

        with freeze_time(records[0]['when'] + datetime.timedelta(seconds=300)):
            client = Mock()
            reminders._scheduled = set()
            reminders.init_reminders(client)
            assert 1234567890 in reminders._scheduled
            # It's 300 seconds, late. Should be 1 day from that point
            reactor.callLater.assert_called_with(86400 - 300, reminders._do_reminder, 1234567890, client)
            db.reminders.save.assert_called_with({
                'hash': 'foo',
                '_id': 1234567890,
                'when': datetime.datetime(day=14, month=12, year=2013),
                'repeat': range(7),
            })


class NextOccurrenceTestCase(TestCase):

    def test_next_occurrence(self):
        reminder = {
            '_id': 1,
            # Start on a wednesday
            'when': datetime.datetime(day=13, month=8, year=2014),
        }

        past = (0, 5)  # Expect 5 days in the future
        today = (2, 7)  # Expect 1 week in the future
        future = (5, 3)  # Expect 3 days in the future

        with freeze_time(reminder['when']):
            for repeat, expect_delta in (past, today, future):
                reminder['repeat'] = [repeat]
                next_time, next_delta = reminders.next_occurrence(reminder)
                assert expect_delta == next_delta
                assert next_time == reminder['when'] + datetime.timedelta(days=expect_delta)

    @patch('helga.plugins.reminders._scheduled')
    @patch('__builtin__.next')
    def test_when_no_next_dow(self, _next, scheduled):
        _next.side_effect = StopIteration

        reminder = {
            '_id': 1,
            'when': datetime.datetime(day=13, month=8, year=2014),
            'repeat': range(7),
        }

        assert reminders.next_occurrence(reminder) is None
        scheduled.discard.assert_called_with(1)


class DeleteReminderTestCase(TestCase):

    @patch('helga.plugins.reminders.db')
    def test_no_found_record(self, db):
        db.reminders.find_one.return_value = None
        retval = reminders.delete_reminder('#bots', 'foo')
        assert retval == "No reminder found with hash 'foo'"

    @patch('helga.plugins.reminders.db')
    def test_deletes_record(self, db):
        db.reminders.find_one.return_value = {'_id': 1}
        reminders.delete_reminder('#bots', 'foo')
        db.reminders.remove.assert_called_with(1)


class ReminderSubcommandTestCase(TestCase):

    @patch('helga.plugins.reminders.in_reminder')
    def test_in_reminder(self, in_reminder):
        client = Mock()
        reminders.reminders(client, '#bots', 'me', 'message', 'in', ['args'])
        in_reminder.assert_called_with(client, '#bots', 'me', ['args'])

    @patch('helga.plugins.reminders.at_reminder')
    def test_at_reminder(self, at_reminder):
        client = Mock()
        reminders.reminders(client, '#bots', 'me', 'message', 'at', ['args'])
        at_reminder.assert_called_with(client, '#bots', 'me', ['args'])

    @patch('helga.plugins.reminders.list_reminders')
    def test_list_reminders(self, list_reminder):
        client = Mock()
        reminders.reminders(client, '#bots', 'me', 'message', 'reminders', ['list'])
        list_reminder.assert_called_with(client, 'me', '#bots')
        client.me.assert_called_with('#bots', 'whispers to me')

    @patch('helga.plugins.reminders.list_reminders')
    def test_list_reminders_with_channel(self, list_reminder):
        client = Mock()
        reminders.reminders(client, '#bots', 'me', 'message', 'reminders', ['list', '#blah'])
        list_reminder.assert_called_with(client, 'me', '#blah')
        client.me.assert_called_with('#bots', 'whispers to me')

    @patch('helga.plugins.reminders.delete_reminder')
    def test_delete_reminder(self, delete_reminder):
        client = Mock()
        reminders.reminders(client, '#bots', 'me', 'message', 'reminders', ['delete', '1'])
        delete_reminder.assert_called_with('#bots', '1')



# -*- coding: utf8 -*-
import re

from mock import Mock, call, patch
from unittest import TestCase

from helga import comm


class FactoryTestCase(TestCase):

    def setUp(self):
        self.factory = comm.Factory()

    def test_build_protocol(self):
        client = self.factory.buildProtocol('address')
        assert client.factory == self.factory

    @patch('helga.comm.settings')
    @patch('helga.comm.reactor')
    def test_client_connection_lost_retries(self, reactor, settings):
        settings.AUTO_RECONNECT = True
        settings.AUTO_RECONNECT_DELAY = 1
        connector = Mock()
        self.factory.clientConnectionLost(connector, Exception)
        reactor.callLater.assert_called_with(1, connector.connect)

    @patch('helga.comm.settings')
    def test_client_connection_lost_raises(self, settings):
        settings.AUTO_RECONNECT = False
        connector = Mock()
        self.assertRaises(Exception, self.factory.clientConnectionLost, connector, Exception)

    @patch('helga.comm.settings')
    @patch('helga.comm.reactor')
    def test_client_connection_failed(self, reactor, settings):
        settings.AUTO_RECONNECT = False
        self.factory.clientConnectionFailed(Mock(), reactor)
        assert reactor.stop.called

    @patch('helga.comm.settings')
    @patch('helga.comm.reactor')
    def test_client_connection_failed_retries(self, reactor, settings):
        settings.AUTO_RECONNECT = True
        settings.AUTO_RECONNECT_DELAY = 1
        connector = Mock()
        self.factory.clientConnectionFailed(connector, reactor)
        reactor.callLater.assert_called_with(1, connector.connect)


class ClientTestCase(TestCase):

    def setUp(self):
        self.client = comm.Client()

    def test_parse_nick(self):
        nick = self.client.parse_nick('foo!~foobar@localhost')
        assert nick == 'foo'

    def test_parse_nick_unicode(self):
        nick = self.client.parse_nick(u'☃!~foobar@localhost')
        assert nick == u'☃'

    @patch('helga.comm.irc.IRCClient')
    def test_me_converts_from_unicode(self, irc):
        snowman = u'☃'
        bytes = '\xe2\x98\x83'
        self.client.me('#foo', snowman)
        irc.describe.assert_called_with(self.client, '#foo', bytes)

    @patch('helga.comm.irc.IRCClient')
    def test_msg_sends_byte_string(self, irc):
        snowman = u'☃'
        bytes = '\xe2\x98\x83'

        self.client.msg('#foo', snowman)
        irc.msg.assert_called_with(self.client, '#foo', bytes)

    def test_alterCollidedNick(self):
        self.client.alterCollidedNick('foo')
        assert re.match(r'^foo_[\d]+$', self.client.nickname)

        # Should take the first part up to '_'
        self.client.alterCollidedNick('foo_bar')
        assert re.match(r'^foo_[\d]+$', self.client.nickname)

    def test_erroneousNickFallback(self):
        assert re.match(r'^helga_[\d]+$', self.client.erroneousNickFallback)

    @patch('helga.comm.settings')
    @patch('helga.comm.smokesignal')
    def test_signedOn(self, signal, settings):
        snowman = u'☃'

        settings.CHANNELS = [
            ('#bots',),
            ('#foo', 'bar'),
            (u'#baz', snowman),  # Handles unicode gracefully?
            ['#a', 'b'],  # As a list
            '#test',  # Single channel
        ]

        with patch.object(self.client, 'join') as join:
            self.client.signedOn()
            assert join.call_args_list == [
                call('#bots'),
                call('#foo', 'bar'),
                call('#baz', snowman),
                call('#a', 'b'),
                call('#test'),
            ]

    @patch('helga.comm.settings')
    @patch('helga.comm.smokesignal')
    def test_signedOn_sends_signal(self, signal, settings):
        settings.CHANNELS = []
        self.client.signedOn()
        signal.emit.assert_called_with('signon', self.client)

    @patch('helga.comm.registry')
    def test_privmsg_sends_single_string(self, registry):
        self.client.msg = Mock()
        registry.process.return_value = ['line1', 'line2']

        self.client.privmsg('foo!~bar@baz', '#bots', 'this is the input')

        args = self.client.msg.call_args[0]
        assert args[0] == '#bots'
        assert args[1] == 'line1\nline2'

    @patch('helga.comm.registry')
    def test_privmsg_responds_to_user_when_private(self, registry):
        self.client.nickname = 'helga'
        self.client.msg = Mock()
        registry.process.return_value = ['line1', 'line2']

        self.client.privmsg('foo!~bar@baz', 'helga', 'this is the input')

        assert self.client.msg.call_args[0][0] == 'foo'

    @patch('helga.comm.settings')
    @patch('helga.comm.irc.IRCClient')
    def test_connectionMade(self, irc, settings):
        self.client.connectionMade()
        irc.connectionMade.assert_called_with(self.client)

    @patch('helga.comm.settings')
    @patch('helga.comm.irc.IRCClient')
    def test_connectionLost(self, irc, settings):
        self.client.connectionLost('an error...')
        irc.connectionLost.assert_called_with(self.client, 'an error...')

    @patch('helga.comm.settings')
    @patch('helga.comm.irc.IRCClient')
    def test_connectionLost_handles_unicode(self, irc, settings):
        snowman = u'☃'
        bytes = '\xe2\x98\x83'
        self.client.connectionLost(snowman)
        irc.connectionLost.assert_called_with(self.client, bytes)

    @patch('helga.comm.smokesignal')
    def test_joined(self, signal):
        # Test str and unicode
        for channel in ('foo', u'☃'):
            assert channel not in self.client.channels
            self.client.joined(channel)
            assert channel in self.client.channels
            signal.emit.assert_called_with('join', self.client, channel)

    @patch('helga.comm.smokesignal')
    def test_left(self, signal):
        # Test str and unicode
        for channel in ('foo', u'☃'):
            self.client.channels.add(channel)
            self.client.left(channel)
            assert channel not in self.client.channels
            signal.emit.assert_called_with('left', self.client, channel)

    def test_kickedFrom(self):
        # Test str and unicode
        for channel in ('foo', u'☃'):
            self.client.channels.add(channel)
            self.client.kickedFrom(channel, 'me', 'no bots allowed')
            assert channel not in self.client.channels

    def test_on_invite(self):
        with patch.object(self.client, 'join') as join:
            self.client.nickname = 'helga'
            self.client.on_invite('me', 'helga', '#bots')
            assert join.called

    def test_on_invite_ignores_other_invites(self):
        with patch.object(self.client, 'join') as join:
            self.client.nickname = 'helga'
            self.client.on_invite('me', 'someone_else', '#bots')
            assert not join.called

    def test_irc_unknown(self):
        with patch.object(self.client, 'on_invite') as on_invite:
            self.client.irc_unknown('me', 'INVITE', ['helga', '#bots'])
            on_invite.assert_called_with('me', 'helga', '#bots')

            on_invite.reset_mock()
            self.client.irc_unknown('me', 'SOME_COMMAND', [])
            assert not on_invite.called

    @patch('helga.comm.smokesignal')
    def test_userJoined(self, signal):
        user = 'helga!helgabot@127.0.0.1'
        self.client.userJoined(user, '#bots')
        signal.emit.assert_called_with('user_joined', self.client, 'helga', '#bots')

    @patch('helga.comm.smokesignal')
    def test_userLeft(self, signal):
        user = 'helga!helgabot@127.0.0.1'
        self.client.userLeft(user, '#bots')
        signal.emit.assert_called_with('user_left', self.client, 'helga', '#bots')

    @patch('helga.comm.irc.IRCClient')
    def test_join_converts_from_unicode(self, irc):
        snowman = u'☃'
        bytes = '\xe2\x98\x83'
        self.client.join(snowman, snowman)
        irc.join.assert_called_with(self.client, bytes, key=bytes)

    @patch('helga.comm.irc.IRCClient')
    def test_leave_converts_from_unicode(self, irc):
        snowman = u'☃'
        bytes = '\xe2\x98\x83'
        self.client.leave(snowman, snowman)
        irc.leave.assert_called_with(self.client, bytes, reason=bytes)

    @patch('helga.comm.log')
    def test_get_channel_logger_no_existing_logger(self, log):
        self.client.channel_loggers = {}
        log.get_channel_logger.return_value = 'foo'

        assert 'foo' == self.client.get_channel_logger('#foo')
        assert '#foo' in self.client.channel_loggers

    @patch('helga.comm.log')
    def test_get_channel_logger_existing_logger(self, log):
        self.client.channel_loggers = {'#foo': 'bar'}
        log.get_channel_logger.return_value = 'foo'

        assert 'bar' == self.client.get_channel_logger('#foo')
        assert not log.get_channel_logger.called

    @patch('helga.comm.settings')
    def test_log_channel_message(self, settings):
        settings.CHANNEL_LOGGING = True
        logger = Mock()

        with patch.object(self.client, 'get_channel_logger'):
            self.client.get_channel_logger.return_value = logger
            self.client.log_channel_message('foo', 'bar', 'baz')
            self.client.get_channel_logger.assert_called_with('foo')
            logger.info.assert_called_with('baz', extra={'nick': 'bar'})



"""
Logging utilities for helga
"""
import datetime
import logging
import logging.handlers
import os
import sys

from helga import settings


def getLogger(name):
    """
    Obtains a named logger and ensures that it is configured according to helga's log settings
    (see :ref:`helga.settings.logging`). Use of this is generally intended to mimic
    :func:`logging.getLogger` with the exception that it takes care of formatters and handlers.

    :param str name: The name of the logger to get
    """
    level = settings.LOG_LEVEL

    logger = logging.getLogger(name)
    logger.setLevel(getattr(logging, level, logging.INFO))
    logger.propagate = False

    # Setup the default handler
    if settings.LOG_FILE:
        handler = logging.handlers.RotatingFileHandler(filename=settings.LOG_FILE,
                                                       maxBytes=50*1024*1024,
                                                       backupCount=6)
    else:
        handler = logging.StreamHandler()
        handler.stream = sys.stdout

    # Setup formatting
    default_format = '%(asctime)-15s [%(levelname)s] [%(name)s:%(lineno)d]: %(message)s'
    formatter = logging.Formatter(settings.LOG_FORMAT or default_format)

    handler.setFormatter(formatter)
    logger.addHandler(handler)

    return logger


def get_channel_logger(channel):
    """
    Obtains a python logger configured to operate as a channel logger.

    :param str channel: the channel name for the desired logger
    """
    logger = logging.getLogger(u'channel_logger/{0}'.format(channel))
    logger.setLevel(logging.INFO)
    logger.propagate = False
    logger.addFilter(UTCTimeLogFilter())

    # Channel logs are grouped into directories by channel name
    log_dir = os.path.join(settings.CHANNEL_LOGGING_DIR, channel)
    if not os.path.exists(log_dir):
        os.makedirs(log_dir)

    # Setup a daily rotating file handler
    handler = ChannelLogFileHandler(log_dir)
    handler.setFormatter(logging.Formatter(u'%(utctime)s - %(nick)s - %(message)s'))
    logger.addHandler(handler)

    return logger


class UTCTimeLogFilter(logging.Filter):
    """
    A log record filter that will add an attribute ``utcnow`` and ``utctime``
    to a log record. The former is a utcnow datetime object, the latter is
    the formatted time of day for utcnow.
    """

    def filter(self, record):
        """
        Filter the log record and add two attributes:

        * ``utcnow``: the value of :func:`datetime.datetime.utcnow`
        * ``utctime``: the time formatted string of ``utcnow`` in the form ``HH:MM:SS``
        """
        record.utcnow = datetime.datetime.utcnow()
        record.utctime = record.utcnow.strftime('%H:%M:%S')
        return True


class ChannelLogFileHandler(logging.handlers.BaseRotatingHandler):
    """
    A rotating file handler implementation that will create UTC dated log files
    suitable for channel logging.
    """

    def __init__(self, basedir):
        """
        :param basedir: The base directory where logs should be stored
        """
        self.basedir = basedir
        filename = os.path.join(basedir, self.current_filename())
        self.next_rollover = self.compute_next_rollover()
        try:
            super(logging.handlers.BaseRotatingHandler, self).__init__(filename, 'a')
        except TypeError:  # pragma: no cover Python >= 2.7
            # python 2.6 uses old-style classes for logging.Handler
            logging.handlers.BaseRotatingHandler.__init__(self, filename, 'a')

    def compute_next_rollover(self):
        """
        Based on UTC now, computes the next rollover date, which will be 00:00:00
        of the following day. For example, if the current datetime is 2014-10-31 08:15:00,
        then the next rollover will be 2014-11-01 00:00:00.
        """
        now = datetime.datetime.utcnow()
        tomorrow = now + datetime.timedelta(days=1)
        return datetime.datetime(tomorrow.year, tomorrow.month, tomorrow.day)

    def current_filename(self):
        """
        Returns a UTC dated filename suitable as a log file. Example: 2014-12-15.txt
        """
        return datetime.datetime.utcnow().strftime('%Y-%m-%d.txt')

    def shouldRollover(self, record):
        """
        Returns True if the current UTC datetime occurs on or after the
        next scheduled rollover datetime. False otherwise.

        :param record: a python log record
        """
        return datetime.datetime.utcnow() >= self.next_rollover

    def doRollover(self):
        """
        Perform log rollover. Closes any open stream, sets a new log filename,
        and computes the next rollover time.
        """
        if self.stream:
            self.stream.close()
            self.stream = None

        self.baseFilename = os.path.abspath(os.path.join(self.basedir, self.current_filename()))
        self.stream = self._open()

        self.next_rollover = self.compute_next_rollover()



from mock import patch

from helga.plugins import reviewboard


@patch('helga.plugins.reviewboard.settings')
def test_reviewboard(settings):
    settings.REVIEWBOARD_URL = 'http://example.com/{review}'
    expected = 'me might be talking about codereview: http://example.com/1234'
    assert reviewboard.reviewboard(None, '#bots', 'me', 'cr1234', ['1234']) == expected


@patch('helga.plugins.reviewboard.settings')
def test_reviewboard_handles_many(settings):
    settings.REVIEWBOARD_URL = 'http://example.com/{review}'
    expected = 'me might be talking about codereview: http://example.com/123, http://example.com/456'
    assert reviewboard.reviewboard(None, '#bots', 'me', 'cr123 cr456', ['123', '456']) == expected



"""
Default settings and configuration utilities
"""
import os
import sys

#: Dictionary of IRC connection details. At a minimum this should contain keys
#: ``HOST`` and ``PORT`` which default to 'localhost' and 6667 respectively.
#: Optionally, you can specify a boolean key ``SSL`` if you require helga to
#: connect to IRC via SSL. You may also specify keys ``USERNAME`` and ``PASSWORD``
#: if your IRC server requires authentication. For example::
#:
#:     SERVER = {
#:         'HOST': 'localhost',
#:         'PORT': 6667,
#:         'SSL': False,
#:         'USERNAME': 'user',
#:         'PASSWORD': 'pass',
#:     }
SERVER = {
    'HOST': 'localhost',
    'PORT': 6667,
}


#: A string for the logging level helga should use for process logging
LOG_LEVEL = 'DEBUG'

#: A string, if set, a string indicating the log file for python logs. By default helga
#: will log directly to stdout
LOG_FILE = None

#: A string that is compatible with configuring a python logging formatter.
LOG_FORMAT = '%(asctime)-15s [%(levelname)s] [%(name)s:%(lineno)d]: %(message)s'

#: Integer value for 'low' priority plugins (see :ref:`plugins.priorities`)
PLUGIN_PRIORITY_LOW = 25

#: Integer value for 'normal' priority plugins (see :ref:`plugins.priorities`)
PLUGIN_PRIORITY_NORMAL = 50

#: Integer value for 'high' priority plugins (see :ref:`plugins.priorities`)
PLUGIN_PRIORITY_HIGH = 75

#: A boolean, if True, will enable conversation logging on all channels
CHANNEL_LOGGING = False

#: If :data:`CHANNEL_LOGGING` is enabled, this is a string of the directory to which channel logs
#: should be written.
CHANNEL_LOGGING_DIR = '.logs'

#: A list of channel names (either with or without a '#' prefix) that will be hidden in the
#: browsable channel log web ui.
CHANNEL_LOGGING_HIDE_CHANNELS = []

#: The preferred nick of the bot instance
NICK = 'helga'

#: A list of channels to automatically join. You can specify either a single channel name
#: or a two-tuple of channel name, and password. For example::
#:
#:     CHANNELS = [
#:         '#bots',
#:         ('#foo', 'password'),
#:     ]
#:
#: Note that this setting is only for hardcoded autojoined channels. Helga also responds
#: to /INVITE commands as well offers a builtin plugin to configure autojoin channels at
#: runtime (see :ref:`builtin.plugins.operator`)
CHANNELS = [
    ('#bots',),
]

#: A boolean indicating if the bot automatically reconnect on connection lost
AUTO_RECONNECT = True

#: An integer for the time, in seconds, to delay between reconnect attempts
AUTO_RECONNECT_DELAY = 5

#: An integer indicating the rate limit, in seconds, for messages sent over IRC. This may help
#: to prevent flood, but may degrade the performance of the bot, as it applies to every message
#: sent to IRC.
RATE_LIMIT = None

#: A list of IRC nicks that should be considered operators/administrators
OPERATORS = []

#: A dictionary containing connection info for MongoDB. The minimum settings that should
#: exist here are 'HOST', the MongoDB host, 'PORT, the MongoDB port, and 'DB' which should be the
#: MongoDB collection to use. These values default to 'localhost', 27017, and 'helga' respectively.
#: Both 'USERNAME' and 'PASSWORD' can be specified if MongoDB requires authentication. For example::
#:
#:     DATABASE = {
#:         'HOST': 'localhost',
#:         'PORT': 27017,
#:         'DB': 'helga',
#:         'USERNAME': 'foo',
#:         'PASSWORD': 'bar',
#:     }
DATABASE = {
    'HOST': 'localhost',
    'PORT': 27017,
    'DB': 'helga',
}

#: The default timezone for the bot instance
TIMEZONE = 'US/Eastern'

#: A list of plugin names that should be enabled automatically for any channel. Note that this
#: does not mean plugins that are loaded. By default, any plugin that has been installed will
#: be loaded and made available. This should be a list of the entry point names defined by each
#: plugin. See :ref:`plugins` for more information.
ENABLED_PLUGINS = [
    'dubstep',
    'facts',
    'help',
    'jira',
    'loljava',
    'manager',
    'meant_to_say',
    'oneliner',
    'operator',
    'poems',
    'reminders',
    'reviewboard',
    'stfu',
    'webhooks',
    'wiki_whois',

    # Sometimes, giphy may give back a gif of questionable content
    # 'giphy',

    # These can get super annoying in public channels
    # 'icanhazascii',

    # Generally, olga isn't being used
    # 'no_more_olga',
]

#: A list of webhook names that should be enabled on process startup. If this value is None,
#: then all webhooks available are loaded and made available. An empty list implies that no
#: webhooks will be made available. See :ref:`webhooks` for more details.
ENABLED_WEBHOOKS = None

#: A boolean, if True, the first response received from a plugin will be the only message
#: sent over IRC. If False, all responses are sent.
PLUGIN_FIRST_RESPONDER_ONLY = True

#: A boolean, if True, command plugins can be run by asking directly, such as 'helga foo_command'.
COMMAND_PREFIX_BOTNICK = True

#: A string char, if non-empty, that can be used to invoke a command without requiring the bot's nick.
#: For example 'helga foo' could be run with '!foo'.
COMMAND_PREFIX_CHAR = '!'

#: A boolean that controls the behavior of argument parsing for command plugins. If False,
#: command plugin arguments are parsed using a naive whitespace split. If True, they will
#: be parsed using :func:`shlex.split`. See :ref:`plugins.creating.commands` for more information.
#: The default is False, but this shlex parsing will be the only supported means of argument
#: string parsing in a future version.
COMMAND_ARGS_SHLEX = False

#: A boolean on whether commands should be treated with case insensitivity. For example,
#: a command 'foo' will respond to 'FOO', 'Foo', 'foo', etc.
COMMAND_IGNORECASE = False

# MISC PLUGIN SETTINGS
FACTS_REQUIRE_NICKNAME = False

# Jira settings. If JIRA_SHOW_FULL_DESCRIPTION is false, only links to the Jira
# ticket will be shown. Otherwise, the ticket title will be pulled and shown.
# Full descriptions require JIRA_REST_API to be set.
# JIRA_USERNAME and JIRA_PASSWORD are optional if authentication is required
JIRA_URL = 'http://localhost/{ticket}'
JIRA_REST_API = ''
JIRA_SHOW_FULL_DESCRIPTION = False
JIRA_AUTH = ('', '')

REVIEWBOARD_URL = 'http://localhost/{review}'
WIKI_URL = 'http://localhost/{user}'

#: The integer port the webhooks plugin should listen for http requests.
WEBHOOKS_PORT = 8080

#: List of two-tuple username and passwords used for http webhook basic authentication
WEBHOOKS_CREDENTIALS = []  # Tuples of (user, pass)


def configure(overrides):
    """
    Applies custom configuration to global helga settings. Overrides can either be
    a python import path string like 'foo.bar.baz' or a filesystem path like
    'foo/bar/baz.py'

    :param str overrides: an importable python path string like 'foo.bar' or a filesystem path
                          to a python file like 'foo/bar.py'
    """
    this = sys.modules[__name__]

    # Filesystem path to settings file
    if os.path.isfile(overrides):
        execfile(overrides, this.__dict__)
        return

    # Module import path settings file
    fromlist = [overrides.split('.')[-1]]
    overrides = __import__(overrides, this.__dict__, {}, fromlist)

    for attr in filter(lambda x: not x.startswith('_'), dir(overrides)):
        setattr(this, attr, getattr(overrides, attr))



"""
Utilities for working with unicode and/or byte strings
"""
from decorator import decorator


def to_unicode(bytestr, errors='ignore'):
    """
    Safely convert a byte string to unicode by first checking if it already is unicode before decoding.
    This function assumes UTF-8 for byte strings and by default will ignore any decoding errors.

    :param bytestr: either a byte string or unicode string
    :param str errors: a string indicating how decoding errors should be handled
                       (i.e. 'strict', 'ignore', 'replace')
    """
    if isinstance(bytestr, unicode):
        return bytestr
    return bytestr.decode('utf-8', errors)


def from_unicode(unistr, errors='ignore'):
    """
    Safely convert unicode to a byte string by first checking if it already is a byte string before
    encoding. This function assumes UTF-8 for byte strings and by default will ignore any encoding errors.

    :param unistr: either unicode or a byte string
    :param str errors: a string indicating how encoding errors should be handled
                       (i.e. 'strict', 'ignore', 'replace')
    """
    if not isinstance(unistr, unicode):
        return unistr
    return unistr.encode('utf-8', errors)


@decorator
def to_unicode_args(fn, *args, **kwargs):
    """
    Decorator used to safely convert a function's positional arguments from byte strings to unicode
    """
    args = list(args)
    for i, val in enumerate(args):
        if isinstance(val, str):
            args[i] = to_unicode(val)
    return fn(*args, **kwargs)


@decorator
def from_unicode_args(fn, *args, **kwargs):
    """
    Decorator used to safely convert a function's positional arguments from unicode to byte strings
    """
    args = list(args)
    for i, val in enumerate(args):
        if isinstance(val, unicode):
            args[i] = from_unicode(val)
    return fn(*args, **kwargs)
